import numpy

from dune.perftool.generation import (backend,
                                      cached,
                                      class_member,
                                      domain,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      include_file,
                                      instruction,
                                      preamble,
                                      temporary_variable,
                                      valuearg,
                                      )
from dune.perftool.pdelab.localoperator import lop_template_range_field
from dune.perftool.options import get_option
from dune.perftool.ufl.modified_terminals import Restriction


@preamble
def define_quadrature_rule(name):
    include_file('dune/pdelab/common/quadraturerules.hh', filetag='operatorfile')
    from dune.perftool.pdelab.geometry import name_geometry
    geo = name_geometry()
    order = name_quadrature_order()
    return "const auto {} = quadratureRule({}, {});".format(name, geo, order)


def name_quadrature_rule():
    name = "quadrature_rule"
    define_quadrature_rule(name)
    return name


@preamble
def define_quadrature_bound(name):
    quad_rule = name_quadrature_rule()
    return "auto {} = {}.size();".format(name, quad_rule)


def name_quadrature_bound():
    name = "quadrature_size"
    define_quadrature_bound(name)

    # Quadrature bound is a valuearg for loopy
    valuearg(name, dtype=numpy.int32)

    return name


@iname
def quadrature_iname():
    domain("q", name_quadrature_bound())
    return "q"


@backend(interface="quadinames")
def quadrature_inames():
    return frozenset({quadrature_iname()})


def quadrature_preamble(code, **kw):
    return instruction(inames=quadrature_iname(), code=code, **kw)


def name_quadrature_point():
    return "qp"


def _local_dim():
    # To determine the shape, I do query global information here for lack of good alternatives
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")
    from dune.perftool.pdelab.geometry import name_dimension, name_intersection_dimension
    if it == 'cell':
        dim = name_dimension()
    else:
        dim = name_intersection_dimension()

    return dim


@preamble
def fill_quadrature_points_cache(name):
    from dune.perftool.pdelab.geometry import name_geometry
    geo = name_geometry()
    quad_order = quadrature_order()
    include_file("dune/perftool/localbasiscache.hh", filetag='operatorfile')
    return "fillQuadraturePointsCache({}, {}, {});".format(geo, quad_order, name)


@class_member("operator")
def typedef_quadrature_points(name):
    range_field = lop_template_range_field()
    dim = _local_dim()
    return "using {} = typename Dune::QuadraturePoint<{}, {}>::Vector;".format(name, range_field, dim)


def type_quadrature_points(name):
    name = name.upper()
    typedef_quadrature_points(name)
    return name


@class_member("operator")
def define_quadrature_points(name):
    qp_type = type_quadrature_points(name)
    return "mutable std::vector<{}> {};".format(qp_type, name)


def name_quadrature_points():
    """Name of vector storing quadrature points as class member"""
    dim = _local_dim()
    name = "qp_" + str(dim)
    define_quadrature_points(name)
    fill_quadrature_points_cache(name)
    return name


def define_quadrature_position(name):
    # TODO
    # qp = name_quadrature_point()
    quad_points = name_quadrature_points()
    quad_iname = quadrature_iname()
    return quadrature_preamble(code="{} = {}[{}];".format(name, quad_points, quad_iname),
                               assignees=frozenset({name}),
                               )


def name_quadrature_position():
    name = "pos"
    # To determine the shape, I do query global information here for lack of good alternatives
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")
    from dune.perftool.pdelab.geometry import name_dimension, name_intersection_dimension
    if it == 'cell':
        shape = (name_dimension(),)
    else:
        shape = (name_intersection_dimension(),)
    temporary_variable(name, shape=shape, shape_impl=("fv",))
    define_quadrature_position(name)
    return name


def name_quadrature_position_in_cell(restriction):
    if restriction == Restriction.NONE:
        return name_quadrature_position()
    else:
        from dune.perftool.pdelab.geometry import to_cell_coordinates
        return to_cell_coordinates(name_quadrature_position(), name_quadrature_point(), restriction)


@preamble
def fill_quadrature_weights_cache(name):
    from dune.perftool.pdelab.geometry import name_geometry
    geo = name_geometry()
    quad_order = quadrature_order()
    include_file("dune/perftool/localbasiscache.hh", filetag='operatorfile')
    return "fillQuadratureWeightsCache({}, {}, {});".format(geo, quad_order, name)


@class_member("operator")
def typedef_quadrature_weights(name):
    range_field = lop_template_range_field()
    dim = _local_dim()
    return "using {} = typename Dune::QuadraturePoint<{}, {}>::Field;".format(name, range_field, dim)


def type_quadrature_weights(name):
    name = name.upper()
    typedef_quadrature_weights(name)
    return name


@class_member("operator")
def define_quadrature_weights(name):
    qw_type = type_quadrature_weights(name)
    return "mutable std::vector<{}> {};".format(qw_type, name)


def name_quadrature_weights():
    """"Name of vector storing quadrature weights as class member"""
    dim = _local_dim()
    name = "qw_" + str(dim)
    define_quadrature_weights(name)
    fill_quadrature_weights_cache(name)

    # Quadrature weighs is a globar argument for loopy
    shape = name_quadrature_bound()
    globalarg(name, shape=(shape,), dtype=numpy.float64)

    return name


def _estimate_quadrature_order():
    """Estimate quadrature order using polynomial degree estimation from UFL"""

    # According to UFL documentation estimate_total_polynomial_degree
    # should only be called on preprocessed forms.
    form = get_global_context_value("formdata").preprocessed_form

    # Estimate polynomial degree of integrals of current type (eg 'Cell')
    integral_type = get_global_context_value("integral_type")
    integrals = form.integrals_by_type(integral_type)
    polynomial_degree = 0
    for i in integrals:
        polynomial_degree = max(polynomial_degree, i.metadata()['estimated_polynomial_degree'])

    return polynomial_degree


def quadrature_order():
    """Get quadrature order

    Note: In PDELab quadrature order m means that integration of
    polynomials of degree m is excat.
    """
    if get_option("quadrature_order"):
        quadrature_order = get_option("quadrature_order")
    else:
        quadrature_order = _estimate_quadrature_order()

    return quadrature_order


def name_quadrature_order():
    # Quadrature order from UFL estimation
    quad_order = quadrature_order()
    return str(quad_order)
